/* mpiutil.ac  -  Utility functions for MPI
 * Copyright (C) 1998, 1999 Free Software Foundation, Inc.
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


#ifdef __KERNEL__

#include <linux/kernel.h>   /* We're doing kernel work */
#define __NO_VERSION__      /* This isn't "the" file 
                             * of the kernel module */
#include <linux/module.h>   /* Specifically, a module */
#include <linux/version.h>   /* Not included by 
                              * module.h because 
                              * of the __NO_VERSION__ */

#include<linux/vmalloc.h>

#else

#include <stdlib.h>

/* 
 * Define the get_random bytes function here, because we don't have it outside the kernel
 */
void get_random_bytes(char *p, int length)
{
  int fd;
  int chread;

  fd=open("/dev/urandom",0);
  if(fd<0) perror("ERROR:  ");
  chread = read(fd, p, length);
  if (chread <0) perror("ERROR: ");
  close(fd);
  return;

}

#endif /* __KERNEL__ */

#include "types.h"
#include "config.h"

#include <stdio.h>
/*#include <string.h>*/
/*#include <assert.h>*/

#include "mpi.h"
#include "mpi-internal.h"
/*#include <memory.h> */
#include "util.h"


/****************
 * Note:  It was a bad idea to use the number of limbs to allocate
 *	  because on a alpha the limbs are large but we normally need
 *	  integers of n bits - So we should chnage this to bits (or bytes).
 *
 *	  But mpi_alloc is used in a lot of places :-)
 */
MPI
mpi_alloc( unsigned nlimbs )

{
    MPI a;

#ifdef __KERNEL__
    a = vmalloc( sizeof *a);
#else
    a = malloc( sizeof *a);
#endif
    a->d = nlimbs? mpi_alloc_limb_space( nlimbs, 0 ) : NULL;

    a->alloced = nlimbs;
    a->nlimbs = 0;
    a->sign = 0;
    a->flags = 1;
    a->nbits = 0;
    return a;

}

void *m_alloc_clear(size_t nlimbs)

{

  void * a;
  
#ifdef __KERNEL__
  a = (void *)vmalloc(nlimbs);
#else
  a = (void *)malloc(nlimbs);
#endif

  if(a) memset(a,0,nlimbs);

  return a;
}

void
mpi_m_check( MPI a )
{
  /* No need to check, since we are inside kernel */
  /*m_check(a);
    m_check(a->d);
  */
    return;
}

MPI
mpi_alloc_secure( unsigned nlimbs )

{
 /* inside the kernel, there is no need to care about security.
   * So. This is just a wrapper to mpi_alloc
   */
  
  /*
    MPI a;
    
    
    a = vmalloc( sizeof *a );
    if(a==NULL)logme(1,"Kmalloc failed at mpi_alloc_secure. No space???\n");
    
    a->d = nlimbs? mpi_alloc_limb_space( nlimbs, 1 ) : NULL;
    
    a->alloced = nlimbs;
    a->flags = 1;
    a->nlimbs = 0;
    a->sign = 0;
    a->nbits = 0;
    return a;
  */

  return mpi_alloc(nlimbs);
}


#if 0
static void *unused_limbs_5;
static void *unused_limbs_32;
static void *unused_limbs_64;
#endif

mpi_ptr_t
mpi_alloc_limb_space( unsigned nlimbs, int secure )
{
    size_t len = nlimbs * sizeof(mpi_limb_t);
    mpi_ptr_t p;

#ifdef __KERNEL__
    p = vmalloc( len );
#else
    p = malloc( len);
#endif

if(p) memset(p,0,len);
    return p;
}

void
mpi_free_limb_space( mpi_ptr_t a )
{
    if( !a )
      {
	return;
      }

    m_free(a);
}


void
mpi_assign_limb_space( MPI a, mpi_ptr_t ap, unsigned nlimbs )
{
    mpi_free_limb_space(a->d);
    a->d = ap;
    a->alloced = nlimbs;
}

size_t
m_size( const void *a )
{
    const byte *p = a;
    size_t n;

    n  = ((byte*)p)[-4];
    n |= ((byte*)p)[-3] << 8;
    n |= ((byte*)p)[-2] << 16;
    return n;
}


void * m_realloc(MPI a, size_t n)
{
  void *b;
  size_t len ;


  len = (a->nlimbs * BYTES_PER_MPI_LIMB);
  if(a->d)
    {
      if( len >= n ) /* we don't shrink for now */
	{
	  return a->d;
	}
      
      b = m_alloc_clear(n);
      memcpy(b, a->d, len );
     m_free(a->d);
    }
  else
    {
#ifdef __KERNEL__
      b=vmalloc(n);
#else
      b=malloc(n);
#endif
    }
  return b;

}

/****************
 * Resize the array of A to NLIMBS. the additional space is cleared
 * (set to 0) [done by m_realloc()]
 */
void
mpi_resize( MPI a, unsigned nlimbs )
{
	if( a==NULL)
	{	
	   a=mpi_alloc(32);
	   a->d=m_alloc_clear( nlimbs * sizeof(mpi_limb_t) );
	   return;
	}
    if( nlimbs <= a->alloced )
	return; /* no need to do it */
    /* Note: a->secure is not used - instead the realloc functions
     * take care of it. Maybe we should drop a->secure completely
     * and rely on a mpi_is_secure function, which would be
     * a wrapper around m_is_secure
     */
    if( a->d )
      {
	a->d = m_realloc(a, nlimbs * sizeof(mpi_limb_t) );
      }
    else
      {
        a->d = m_alloc_clear( nlimbs * sizeof(mpi_limb_t) );
      }
    a->alloced = nlimbs;
}

void
mpi_clear( MPI a )
{
    a->nlimbs = 0;
    a->nbits = 0;
    a->flags = 0;
}


void
mpi_free( MPI a )
{
    if( !a )
	return;

    m_free( a->d );
#ifdef __KERNEL__
    vfree(a);
#else
    free(a);
#endif
}

void
m_free(void * a)

{
#ifdef __KERNEL__
  if(a) vfree(a);
#else
  if(a) free(a);
#endif

}

/* We are inside the kernel, so no need to set_secure! */
void
mpi_set_secure( MPI a )
{
  /*
    mpi_ptr_t ap, bp;
    
    if( (a->flags & 1) )
    return;
    a->flags |= 1;
    ap = a->d;
    if( !a->nlimbs ) {
    my_assert(!ap);
    return;
    }
    bp = mpi_alloc_limb_space( a->nlimbs, 1 );
    MPN_COPY( bp, ap, a->nlimbs );
    a->d = bp;
    mpi_free_limb_space(ap);
  */
  a->flags |= 1;
  return;	
}


MPI
mpi_set_opaque( MPI a, void *p, int len )
{
    if( !a ) {
	a = mpi_alloc(0);
    }

    if( a->flags & 4 )
	m_free( a->d );
    else {
	mpi_free_limb_space(a->d);
    }

    a->d = p;
    a->alloced = 0;
    a->nlimbs = 0;
    a->nbits = len;
    a->flags = 4;
    return a;
}


void *
mpi_get_opaque( MPI a, int *len )
{
    if( len )
	*len = a->nbits;
    return a->d;
}


/****************
 * Note: This copy function should not interpret the MPI
 *	 but copy it transparently.
 */
MPI
mpi_copy( MPI a )
{
    int i;
    MPI b;
    if( a && (a->flags & 4) ) {
      void *p = mpi_alloc( a->nbits );
      memcpy( p, a->d, a->nbits );
      b = mpi_set_opaque( NULL, p, a->nbits );
    }
    else if( a ) {
	b = mpi_alloc( a->nlimbs );
	b->nlimbs = a->nlimbs;
	b->sign = a->sign;
	b->flags  = a->flags;
	b->nbits = a->nbits;
	for(i=0; i < b->nlimbs; i++ )
	    b->d[i] = a->d[i];
    }
    else
      {
	b = NULL;
      }

    return b;
}


/****************
 * This function allocates an MPI which is optimized to hold
 * a value as large as the one given in the arhgument and allocates it
 * with the same flags as A.
 */
MPI
mpi_alloc_like( MPI a )
{
    MPI b;

    if( a && (a->flags & 4) ) {
	void *p = mpi_alloc( a->nbits );
	memcpy( p, a->d, a->nbits );
	b = mpi_set_opaque( NULL, p, a->nbits );
    }
    else if( a ) {
	b = mpi_alloc( a->nlimbs );
	b->nlimbs = 0;
	b->sign = 0;
	b->flags = a->flags;
	b->nbits = 0;
    }
    else
	b = NULL;
    return b;
}


void
mpi_set( MPI w, MPI u)
{
    mpi_ptr_t wp, up;
    mpi_size_t usize = u->nlimbs;
    int usign = u->sign;

    RESIZE_IF_NEEDED(w, usize);
    wp = w->d;
    up = u->d;
    MPN_COPY( wp, up, usize );
    w->nlimbs = usize;
    w->nbits = u->nbits;
    w->flags = u->flags;
    w->sign = usign;

}


void
mpi_set_ui( MPI w, unsigned long u)
{

    RESIZE_IF_NEEDED(w, 1);
    w->d[0] = u;
    w->nlimbs = u? 1:0;
    w->sign = 0;
    w->nbits = 0;
    w->flags = 0;
}


MPI
mpi_alloc_set_ui( unsigned long u)
{
    MPI w = mpi_alloc(1);
    w->d[0] = u;
    w->nlimbs = u? 1:0;
    w->sign = 0;
    return w;
}


void
mpi_swap( MPI a, MPI b)
{
    struct gcry_mpi tmp;

    tmp = *a; *a = *b; *b = tmp;
}

